/* -*- mode: c++ -*- */

#include <cuda.h>
#include <stdio.h>
#include <assert.h>


//TODO

extern __shared__ uint smoutput[];

// from http://forums.nvidia.com/index.php?showtopic=186669
__device__ uint get_smid(void) {
     uint ret;
     asm("mov.u32 %0, %smid;" : "=r"(ret) );
     return ret;
}

__global__ void testfn(clock_t *start_times, clock_t *end_times, uint *work, uint *start, uint *end, uint *smids)
{
  int tid = blockIdx.y * gridDim.x + blockIdx.x;
  tid *= blockDim.x * blockDim.y * blockDim.z;
  tid += threadIdx.z * blockDim.x * blockDim.y + threadIdx.y * blockDim.x + threadIdx.x;

  start_times[tid] = clock64();

  //start[tid] = atomicAdd(start, 1);

  uint m_z = tid;
  uint m_w = tid;
  uint rnd;

  uint work_iter;
  work_iter = work[tid];

  for(int i =0; i < work_iter;i++)
    {
      m_z = 36969 * (m_z & 65535) + (m_z >> 16);
      m_w = 18000 * (m_w & 65535) + (m_w >> 16);
      rnd = (m_z << 16) + m_w;     
    }

  smoutput[0] = rnd;

  smids[tid] = get_smid();
  //end[tid + 1] = atomicAdd(end, 1); // could be start, if we wanted to absolutely order stuff?
  
  end_times[tid] = clock64();
}

void make_work(uint *work, int blocks, uint *multiplier, int nmulti)
{
  int i;
  for(i = 0; i < blocks; i++)
    work[i] = 10000*multiplier[i % nmulti];
}

int read_multi(const char *filename, uint **multipliers)
{
  FILE *f;
  int n, i;

  f = fopen(filename, "r");

  assert(f != NULL);

  fscanf(f, "%d", &n);

  assert(n > 0);

  *multipliers = (uint *) calloc(n, sizeof(uint));

  for(i = 0; i < n; i++)
    fscanf(f, "%d", &((*multipliers)[i]));

  fclose(f);  

  return n;
}

int main(int argc, char *argv[])
{
  int BLOCKS=64, GRID_HEIGHT, GRID_WIDTH;
  int THREADS=1;
  int resident_blocks=2;
  struct cudaDeviceProp devinfo;;

  if(argc < 5)
    {
      printf("Usage: %s grid_width grid_height resident_blocks work_multiplier_file\n", argv[0]);
      return 0;
    }


  cudaGetDeviceProperties(&devinfo, 0);

  resident_blocks = atoi(argv[3]);

  GRID_WIDTH = atoi(argv[1]);
  GRID_HEIGHT = atoi(argv[2]);

  BLOCKS = GRID_HEIGHT * GRID_WIDTH;
  
  dim3 threads(THREADS);

  clock_t start_times[BLOCKS*THREADS];
  clock_t end_times[BLOCKS*THREADS];
  uint start[BLOCKS*THREADS + 1];
  uint smids[BLOCKS*THREADS + 1];
  uint end[BLOCKS*THREADS + 1];
  uint work[BLOCKS*THREADS];
  uint *multipliers;
  int nmulti;

  void *start_times_d;
  void *end_times_d; 
  void *start_d;
  void *smids_d;
  void *end_d;
  void *work_d;

  cudaMalloc(&start_times_d, BLOCKS*THREADS*sizeof(clock_t));
  cudaMalloc(&end_times_d, BLOCKS*THREADS*sizeof(clock_t));
  cudaMalloc(&start_d,(BLOCKS*THREADS+1)*sizeof(uint));
  cudaMalloc(&smids_d,(BLOCKS*THREADS+1)*sizeof(uint));
  cudaMalloc(&end_d,(BLOCKS*THREADS+1)*sizeof(uint));
  cudaMalloc(&work_d,(BLOCKS*THREADS)*sizeof(uint));


  cudaMemset(start_times_d, 0, BLOCKS*THREADS*sizeof(clock_t));
  cudaMemset(end_times_d, 0,BLOCKS*THREADS*sizeof(clock_t));
  cudaMemset(start_d, 0,(BLOCKS*THREADS+1)*sizeof(uint));
  cudaMemset(smids_d, 0,(BLOCKS*THREADS+1)*sizeof(uint));
  cudaMemset(end_d, 0,(BLOCKS*THREADS+1)*sizeof(uint));

  nmulti = read_multi(argv[4], &multipliers);
  make_work(work, BLOCKS*THREADS, multipliers, nmulti);

  cudaMemcpy(work_d, work, BLOCKS*THREADS*sizeof(uint), cudaMemcpyHostToDevice);

  dim3 gd(GRID_WIDTH, GRID_HEIGHT);

  testfn<<<gd,THREADS,(devinfo.sharedMemPerBlock/(resident_blocks+1))+1>>>((clock_t *) start_times_d, 
                                                            (clock_t *) end_times_d, (uint *) work_d,
                                                            (uint *) start_d, (uint *) end_d, (uint *) smids_d);

  cudaMemcpy(start_times, start_times_d, BLOCKS*THREADS*sizeof(clock_t), cudaMemcpyDeviceToHost);
  cudaMemcpy(end_times, end_times_d, BLOCKS*THREADS*sizeof(clock_t), cudaMemcpyDeviceToHost);
  cudaMemcpy(start, start_d, (BLOCKS*THREADS+1)*sizeof(uint), cudaMemcpyDeviceToHost);
  cudaMemcpy(smids, smids_d, (BLOCKS*THREADS+1)*sizeof(uint), cudaMemcpyDeviceToHost);
  cudaMemcpy(end, end_d, (BLOCKS*THREADS+1)*sizeof(uint), cudaMemcpyDeviceToHost);

  printf("##GPU\n%s\n", devinfo.name);
  printf("##GPUSPECS\n%d %d\n", devinfo.multiProcessorCount, devinfo.sharedMemPerBlock);
  printf("##COMPUTECAPABILITY\n%d %d\n", devinfo.major, devinfo.minor);
  printf("##KERNELCFG\n%s %d %d %d %d %d\n", "schedule_block_syn", GRID_HEIGHT, GRID_WIDTH, THREADS, 1, 1);
  printf("##RESIDENT\n%d\n", resident_blocks);

  printf("##THREADS\n");
  for(int i = 0; i < BLOCKS*THREADS; i++)
    {
      printf("%d %lu %lu %u\n", i, start_times[i], end_times[i], smids[i]);
    }
}
