#!/usr/bin/env python

# Author: Sreepathi Pai

import sched
import sys

from cuda_timing import read_header, read_threads_as_block, MPROC, GRID_HEIGHT, GRID_WIDTH, THREADS_MPROC, GBID

def read_timing_file(f):
    ff = open(f, "r")
    data_header, headers = read_header(ff)
    
    MAXBLOCKS=headers["RESIDENT"]*headers["GPUSPECS"][MPROC]*4

    out = [b for b in read_threads_as_block(ff, headers['KERNELCFG'], start_block=0, max_blocks = MAXBLOCKS)]

    return headers, out

headers, blocks = read_timing_file(sys.argv[1])

kg = sched.KernelGrid(headers['KERNELCFG'][GRID_WIDTH], headers['KERNELCFG'][GRID_HEIGHT], headers['RESIDENT'])
gpu = sched.GPUS[headers['GPU']]()

print "Kernel %dx%d blocks residency=%d on '%s' with %d multiprocessors." % (kg.gridX, kg.gridY, kg.residency, headers['GPU'], headers['GPUSPECS'][MPROC])

block_to_sm = dict([(blk[GBID], blk[THREADS_MPROC]) for blk in blocks])

ok = True
for x in gpu.sched_initial(kg):
    order, linblkid, blkid, cluster, mp, priority = x
    # order, blockid, smid, cluster, mp, cluster, priority
    print "%3d  %3d  %2d %2d  %2d %2d %2d " % (order, linblkid, block_to_sm[linblkid], gpu.mp_to_cluster_map[block_to_sm[linblkid]], mp, cluster, priority),
    if mp != block_to_sm[linblkid]:
        print "*"
        ok = False
    else:
        print ""

if ok:
    print "schedule OK"
else:
    print "schedule WRONG"
